Class = require 'lib/Class'
Vector = require 'lib/Vector'
ImageManager = require 'ImageManager'
require 'lib/Util'

local Entity = Class({
	type = "Entity";
	name = "";
	position = Vector(0.0, 0.0);
	scale = Vector(1.0, 1.0);
	angle = 0.0;

	image_filename = "";
	image = nil;

	parent = nil;
	children = {};

	orbiting = false;
	orbit = { };

	ToLocalSpace = function(self, vec)
		local c, s = math.cos(self.angle), math.sin(self.angle)
		local x, y = vec:unpack()
		--Do the translation
		x, y = x - self.x, y - self.y
		--Do the rotation
		x, y = c * x - s * y, s * x + c * y
		--Do the scale 
		--x, y = x * self.scale.x, y * self.scale.y
		return Vector(x, y)
	end;

	ToWorldSpace = function(self, vec)
		local c, s = math.cos(-self.angle), math.sin(-self.angle)
		local x, y = vec:unpack()
		--Undo the scale
		--x, y = x / self.scale.x, y / self.scale.y
		
		--Undo the rotation
		x, y = c*x - s*y, s*x + c*y

		--Undo the translation
		x = x + self.position.x
		y = y + self.position.y

		return Vector(x, y)
	end;

	GetWorldPosition = function(self)
		local pos = self.position
		local par = self.parent
		while par ~= nil do
			pos = par:ToWorldSpace(pos)
			par = par.parent
		end
		return pos
	end;

	SetOrbitHeight = function(self, height)
		self.orbit.height = height
		self.orbit.desired_height = height
	end;

	MoveToOrbitHeight = function(self, height)
		self.orbit.desired_height = height
	end;

	AlignToOrbit = function(self)
		self.angle = self.orbit.current_angle + math.pi
	end;

	GetOrbitalPeriod = function(self)
		return (math.pi*2) / (self.orbit.speed / self.orbit.height)
	end;

	GetAngularSpeed = function(self)
		return (self.orbit.speed / self.orbit.height)
	end;

	SwapOrbit = function(self, new_orbit)
		self:SetParent(new_orbit)
	end;

	Update = function(self, dt)
		for i,v in pairs(self.children) do 
			v:Update(dt)
		end

		if self.orbiting then
			local angular_speed = (self.orbit.speed * dt) / self.orbit.height
			self.orbit.current_angle = normalize_rad(self.orbit.current_angle + angular_speed)

			local height_diff = self.orbit.height - self.orbit.desired_height
			if math.abs(height_diff) > self.orbit.height_change_speed * dt then
				local sign = 0
				if height_diff > 0 then
					sign = -1
				else
					sign = 1
				end
				self.orbit.height = self.orbit.height + (self.orbit.height_change_speed * dt * sign)
			else
				self.orbit.height = self.orbit.desired_height
			end

			self.position = Vector(math.cos(self.orbit.current_angle) * self.orbit.height, 
								  math.sin(self.orbit.current_angle) * self.orbit.height)
		end
	end;

	Draw = function(self)
		--local wx, wy = game.camera:ToCameraCoords(self:GetWorldPosition()):unpack()

		local x, y = self.position:unpack()
		local angle = self.angle
		local sx, sy = self.scale:unpack()
		if self.parent ~= nil then
			x, y = self.parent:ToWorldSpace(Vector(x, y)):unpack()
		end

		love.graphics.push()
		love.graphics.translate(x, y)
		love.graphics.rotate(angle)

		for i,v in pairs(self.children) do
			v:Draw()
		end

		if self.image ~= nil then
			--[[local w, h = self.image:getWidth(), self.image:getHeight()
			if wx < 0 or wx  > love.graphics:getWidth() or
			   wy < 0 or wy  > love.graphics:getHeight() then
			--culling is messed up, just disable for now
			--	 love.graphics.pop()
			--   return
			end]]
			love.graphics.draw(self.image, 0, 0, 0, sx, sy, self.image:getWidth() / 2, self.image:getHeight() / 2)
		end
		love.graphics.pop()
	end;

	SetParent = function(self, parent)
		if self.parent ~= nil then
			for i,v in pairs(self.parent.children) do
				if v == self then
					table.remove(self.parent.children, i)
					break
				end
			end
		end
		self.parent = parent
		self.parent.children[#self.parent.children + 1] = self
	end;

	SetScale = function(self, new_scale)
		self.scale = new_scale
	end;

	init = function(self, name, image_filename, parent)
		self.children = { }
		self.orbit = {
			height = 0;
			desired_height = 0;
			height_change_speed = 0;
			speed = 0;
			current_angle = 0;
		}
		if name then
			self.name = name
		else self.name = "Unnamed" end

		if image_filename then
			self.image_filename = image_filename
			self.image = imagemanager:get(self.image_filename)
		end

		if parent ~= nil then
			self.parent = parent
			self.parent.children[#self.parent.children+1] = self
		end
	end	
})

return Entity